"""Support utilities for ciphers. 

Language models are for understanding what English looks like, for help with 
cipher breaking.

Text prettify is various procedures for pretty-printing text, such as
segmenting text into words.

Utilities are general utility functions.
"""
